const stringBuilder = require('./stringBuilder');

//解析compactArray文本到一个对象。结果在返回的对象的data属性中（其他属性返回一些辅助信息）。
//【该类仅依赖stringBuilder，可以一起剥离。】
function fromCompactArray(compactArrayString) {
    if (typeof (compactArrayString) != 'string')
        return;
    var curIndex = -1;
    var next = function () {
        curIndex++;
        var saveIndex = curIndex;
        var cur;

        var endpos = compactArrayString.length;
        //多数字符串不需要转义，所以这里先优化处理（先判断不需要转义的字符串）
        while (true) {
            if (curIndex >= endpos || compactArrayString[curIndex] == ',')//整段字符段不需要转义，则直接拷贝整段。
                return compactArrayString.slice(saveIndex, curIndex);
            if (compactArrayString[curIndex] == '\\') {
                //在第一次出现转义字符进入转义处理。
                cur = new stringBuilder();
                break;
            }
            curIndex++;
        }

        var ch;
        while (true) {
            if (curIndex >= endpos || compactArrayString[curIndex] == ',') {
                cur.append(compactArrayString.slice(saveIndex, curIndex));
                return cur.toString();
            }
            if (compactArrayString[curIndex] == '\\') {
                cur.append(compactArrayString.slice(saveIndex, curIndex));
                curIndex++;
                ch = compactArrayString[curIndex];
                cur.append(ch == 'c' ? ',' : ch);
                saveIndex = curIndex + 1;
            }
            curIndex++;
        }
    };
    var convert = function (value, toTypeString) {
        if (toTypeString == 'string')
            return value;

        if (toTypeString == 'int')
            return parseInt(value) || 0;
        if (toTypeString == 'long')//大整数存在精度丢失问题，只能改用字符串。参考：http://www.cnblogs.com/snandy/p/4943138.html
            return value || '0';
        if (toTypeString == 'double' || toTypeString == 'decimal')
            return parseFloat(value) || 0;
        if (toTypeString == 'bool')
            return value == '1';
        if (toTypeString == 'date')
            return value && new Date(value);

        return value;
    };

    var r = {};
    r.ver = next(); //要求：@mk:compactArray&ver=1;
    if (!r.ver || r.ver.indexOf('@mk:compactArray') != 0)
        return null;

    var rowCount = parseInt(next());
    var fieldCount = parseInt(next());
    var hasType = next() == '1';
    var zip = next() == '1';

    r.fields = [];
    for (var i = 0; i < fieldCount; i++) {
        r.fields[i] = next();
    }

    if (hasType) {
        r.fieldTypes = [];
        for (var i = 0; i < fieldCount; i++) {
            r.fieldTypes[i] = next();
        }
    }

    var data = [];
    var nextWithZip;
    if (zip) {
        var zipTable = [];
        nextWithZip = function () {
            if (compactArrayString[curIndex + 1] == '#') {
                curIndex++;
                var n = next();
                return zipTable[parseInt(n)];
            }
            else {
                var n = next();
                if (n.length >= 8)
                    zipTable.push(n);
                return n;
            }
        };
    }
    else
        nextWithZip = next;

    for (var i = 0; i < rowCount; i++) {
        var obj = {};
        for (var j = 0; j < fieldCount; j++) {
            obj[r.fields[j]] = hasType ? convert(nextWithZip(), r.fieldTypes[j]) : nextWithZip();
        }
        data[i] = obj;
    }
    r.data = data;

    return r;
}

module.exports = fromCompactArray;