class Random {
    constructor(seed) {
        this.seed = seed;
    }

    /**
     * @param {number} max
     * @returns {number} */
    next(max) {
        this.seed = (this.seed * 9301 + 49297) % 233280; //为何使用这三个数? 参考：https://www.zhihu.com/question/22818104
        let val = this.seed / (233280.0);
        return max ? Math.trunc(val * max) : val;
    }
}

let baseTime = new Date("2013-05-03 15:30:00").getTime(); //以吴怡婷的诞辰时间点为基准
/**@type {Random} */
let rand = new Random(Date.now() - baseTime);

/**
 log(Number.MAX_SAFE_INTEGER.toString().length); //输出：16
 log(Number.MAX_SAFE_INTEGER);//js的最大安全整数（超出后会有精度丢失问题），输出：9 007 199 254 740 991 （9000多万亿）
log(Date.now().toString().length); //输出：13
let date = new Date('9999-12-31 23:59:59.999');
log(date.getTime().toString().length); //输出：15


let date = new Date(2100, 0);
let val = date.getTime() * 1000;
log(val < Number.MAX_SAFE_INTEGER);//输出: true
log(val.fmt()); //输出：4,102,416,000,000,000
log(val.toString().length); //输出：16
log(new Date(val / 1000).fmt('f')); //输出：2100-01-01 00:00:00.0

经测算，最大大约可以安全运行到 "2298-01-01 00:00:00" (产生的id值为8983125000000442， 8983多万亿, 比最大安全整数小)
 */

/**
 * 用当前时间数值（unix时间，表示距1970-01-01零点的秒数， js里为毫秒数）作为新的id
 */
module.exports = {
    get newId() {
        let val = Date.now();
        if (val <= this.lastId)
            val = ++this.lastId;
        else
            this.lastId = val;

        return this.toId(val);
    },

    toId(val){
        return (val - baseTime) * 1000 + rand.next(1000);
    },

    rand(max) {
        return rand.next(max);
    },

    time(id) {
        return getTime(id);
    },

    MIN_ID: 10000 * 10000
}

function getTime(id) {
    return new Date(Math.trunc(id / 1000) + baseTime);
}
