"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const wcl = require("../../jswcl-ts");
const _ = __importStar(require("lodash"));
var RawType;
(function (RawType) {
    RawType["string"] = "string";
    RawType["number"] = "number";
    RawType["date"] = "date";
    RawType["time"] = "time";
    RawType["bool"] = "bit";
    RawType["image"] = "image";
    RawType["guid"] = "guid";
})(RawType = exports.RawType || (exports.RawType = {}));
function type1(name, rawType, options) {
    let obj = new ColType(name, rawType, undefined);
    wcl.extends_(obj, options);
    return obj;
}
function type2(name, refType, options) {
    let obj = new ColType(name, refType.rawType, refType);
    wcl.extends_(obj, options);
    return obj;
}
function getColeType(name) {
    let ret = ColType[name];
    if (!ret)
        ret = _.find(ColType, i => i instanceof ColType && i.name === name);
    return ret;
}
exports.getColeType = getColeType;
class ColType {
    constructor(name, rawType, refType) {
        this.name = '';
        this.rawType = RawType.string;
        this.length = 0;
        this.prec = 0;
        this.name = name;
        this.refType = refType;
        this.isFixLength = false;
        if (refType) {
            this.rawType = refType.rawType;
            this.hasLengthProp = refType.hasLengthProp;
            this.isFixLength = refType.isFixLength;
        }
        else {
            this.rawType = rawType || RawType.string;
            this.hasLengthProp = this.rawType === RawType.string;
            if (name === 'dicimal')
                this.hasLengthProp = true;
            if (name === 'text')
                this.hasLengthProp = false;
            if (name === 'char' || name === 'nchar')
                this.isFixLength = true;
        }
        this.needQuoted = this.rawType !== RawType.number && this.rawType !== RawType.bool;
    }
    get baseType() {
        if (this.refType)
            return this.refType.baseType;
        return this;
    }
    toSqlType(length, prec) {
        let baseType = this.baseType;
        length = length || this.length || baseType.length;
        prec = prec || this.prec || baseType.prec;
        if (length > 0 && prec > 0)
            return `${baseType.name}(${length}, ${prec})`;
        else if (length > 0)
            return `${baseType.name}(${length})`;
        else
            return baseType.name;
    }
}
ColType.char = type1('char');
ColType.varchar = type1('varchar');
ColType.nchar = type1('nchar');
ColType.nvarchar = type1('nvarchar');
ColType.text = type1('text');
ColType.int = type1('int', RawType.number);
ColType.bigint = type1('bigint', RawType.number);
ColType.float = type1('decimal', RawType.number);
ColType.bool = type1('bit', RawType.bool);
ColType.time = type1('datetime', RawType.time);
ColType.image = type1('image', RawType.image);
ColType.guid = type1('uniqueidentifier', RawType.guid);
ColType.date = type2('date', ColType.char, { length: 10 });
ColType.shortTime = type2('shortTime', ColType.char, { length: 5 });
ColType.longId = type2('longId', ColType.bigint);
ColType.userId = type2('userId', ColType.longId);
ColType.id = type2('id', ColType.longId);
ColType.oid = type2('oid', ColType.varchar, { length: 50 });
ColType.xml = type2('xml', ColType.text);
ColType.json = type2('json', ColType.text);
ColType.cdoe1 = type2('c1', ColType.char, { length: 1 });
ColType.cdoe2 = type2('c2', ColType.char, { length: 2 });
ColType.cdoe4 = type2('c4', ColType.char, { length: 4 });
ColType.cdoe8 = type2('c8', ColType.char, { length: 8 });
ColType.code = type2('code', ColType.char, { length: 8 });
ColType.float2 = type2('f2', ColType.float, { length: 19, prec: 2 });
ColType.float4 = type2('f4', ColType.float, { length: 19, prec: 4 });
ColType.float8 = type2('f8', ColType.float, { length: 19, prec: 8 });
ColType.text1 = type2('t1', ColType.varchar, { length: 50 });
ColType.text2 = type2('t2', ColType.varchar, { length: 200 });
ColType.text3 = type2('t3', ColType.varchar, { length: 500 });
ColType.text4 = type2('t4', ColType.varchar, { length: 2000 });
ColType.text5 = type2('t5', ColType.varchar, { length: 8000 });
exports.ColType = ColType;
function col(declare, extension) {
    return Column.parse(undefined, declare, extension);
}
exports.col = col;
exports.sybOption = Symbol('option');
exports.sybIndex = Symbol('index');
function getColsOption(cols) {
    let ret = cols[exports.sybOption];
    if (!ret)
        cols[exports.sybOption] = ret = {};
    return ret;
}
exports.getColsOption = getColsOption;
function makeColumns(fromObj) {
    let obj = fromObj;
    let option = getColsOption(obj);
    if (obj.hasOwnProperty(0)) {
        obj[exports.sybOption] = obj[0];
        delete obj[0];
    }
    if (wcl.isString(option)) {
        option = { tableName: option };
        obj[exports.sybOption] = option;
    }
    if (option.logAble) {
        obj.createTime = exports.CLogger.createTime;
        obj.updateTime = exports.CLogger.updateTime;
    }
    let keys = Object.keys(obj);
    let pk = [];
    let titles = {};
    let lastGrps = [];
    for (const key of keys) {
        const column = obj[key];
        if (wcl.isString(column)) {
            obj[key] = col(column);
        }
        else if (column instanceof ColType) {
            obj[key] = new Column(key, column);
        }
        obj[key].jsName = key;
        if (obj[key].primaryKey)
            pk.push(obj[key]);
        if (!column.fieldName)
            column.fieldName = column.jsName;
        titles[key] = obj[key].caption || obj[key].jsName;
        if (column.index) {
            if (!obj[exports.sybIndex])
                obj[exports.sybIndex] = {};
            obj[exports.sybIndex]['ix_' + column.fieldName] = column.fieldName + (column.unique ? ':uq' : '');
        }
        if (!column.grps)
            column.grps = lastGrps;
        else
            lastGrps = column.grps;
    }
    option.keys = pk;
    option.titles = titles;
    return obj;
}
exports.makeColumns = makeColumns;
class Column {
    constructor(name, type) {
        this.jsName = '';
        this.fieldName = '';
        this.type = ColType.text1;
        this.caption = '';
        this.notnull = false;
        this.unique = false;
        this.index = false;
        this.primaryKey = false;
        this.fieldName = name || '';
        if (type)
            this.type = type;
    }
    toSqlType() {
        return this.type.toSqlType(this.length, this.prec);
    }
    static parse(toColumn, declare, extension) {
        let isNew = !toColumn;
        if (!toColumn)
            toColumn = new Column();
        if (declare) {
            let sr = new wcl.script.reader(declare);
            let curr, name, value;
            if (isNew) {
                curr = sr.popNext();
                if (!curr || !curr.isIdentifyOrString())
                    throw new Error('IColumn.declare 第一节必须是字段类型\r\n' + declare);
                let name = curr.getValue();
                if (name) {
                    toColumn.type = getColeType(name);
                    if (!toColumn.type)
                        throw new Error('IColumn.declare 字段类型非法: ' + name + '\r\n' + declare);
                }
                curr = sr.peekNext();
                if (curr.isArray()) {
                    let value = curr.getValue();
                    if (value.length > 0)
                        toColumn.length = parseInt(value[0]);
                    if (value.length > 1)
                        toColumn.prec = parseInt(value[1]);
                    sr.popNext();
                    curr = sr.peekNext();
                }
                if (curr && !curr.isIdentifyOrString())
                    throw new Error('IColumn.declare 第二节必须是字段标题\r\n' + declare);
                if (curr) {
                    let val = curr.getValue();
                    if (val)
                        toColumn.caption = val;
                    sr.popNext();
                }
            }
            while (true) {
                curr = sr.popNext();
                if (!curr)
                    break;
                if (!curr.isIdentifyOrString())
                    continue;
                name = curr.name;
                if (name === 'notnull')
                    toColumn.notnull = true;
                else if (name === 'pk')
                    toColumn.primaryKey = true;
                else if (name === 'uq')
                    toColumn.unique = true;
                else if (name === 'ix')
                    toColumn.index = true;
                else {
                    value = curr.getValue();
                    if (name === 'field')
                        toColumn.fieldName = value;
                    else if (name === 'default')
                        toColumn.defaultValue = value;
                    else if (name === 'note')
                        toColumn.note = value;
                    else if (name === 'hint')
                        toColumn.hint = value;
                    else if (name === 'grp') {
                        if (value) {
                            value = value.split(',');
                            if (!toColumn.grps)
                                toColumn.grps = value;
                            else
                                toColumn.grps.push(...value);
                        }
                    }
                }
            }
        }
        wcl.extends_(toColumn, extension);
        if (!toColumn.fieldName)
            toColumn.fieldName = toColumn.jsName;
        return toColumn;
    }
    [Symbol.toPrimitive](hint) {
        return this.jsName;
    }
    sqlValue(val) {
        return convertValue(val, this.type, true);
    }
    valuePart(val, withFieldName) {
        val = this.sqlValue(val);
        if (val == undefined)
            val = 'null';
        else if (this.type.needQuoted) {
            if (wcl.isString(val))
                val = val.replace(replaceQuoted, "''");
            val = val.q;
        }
        return withFieldName ? `[${this.fieldName}] = ${val}` : val;
    }
    nameAndCaption() {
        if (!this.caption || this.jsName === this.caption)
            return this.jsName;
        return `${this.jsName}(${this.caption})`;
    }
}
exports.Column = Column;
let replaceQuoted = /'/g;
function convertToDate(val) {
    if (val == undefined)
        return new Date(0);
    if (wcl.isString(val)) {
        return val.toDate();
    }
    return val;
}
exports.convertToDate = convertToDate;
function convertValue(val, toType, finalAsString) {
    if (toType.rawType === RawType.bool)
        return val ? '1' : '0';
    if (toType === ColType.date)
        return Date.prototype.fmt.call(convertToDate(val), 'D') || '';
    if (toType === ColType.time)
        return Date.prototype.fmt.call(convertToDate(val), 'L');
    if (toType.rawType === RawType.string) {
        if (val == undefined)
            return '';
        return String(val);
    }
    if (val != undefined)
        return finalAsString ? String(val) : val;
}
exports.convertValue = convertValue;
function getGroupFields(colsObj, grps) {
    return _.filter(colsObj, i => i.grps && _.some(grps, j => _.some(i.grps, k => k === j))).map(i => i.jsName);
}
exports.getGroupFields = getGroupFields;
exports.CLogger = {
    createTime: col('datetime 创建时间 grp="-"'),
    updateTime: col('datetime 修改时间'),
};
makeColumns(exports.CLogger);
//# sourceMappingURL=column.js.map