"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const wcl = require("../../jswcl-ts");
const cols = __importStar(require("./column"));
const dbApi = __importStar(require("./db"));
exports.sybOwner = Symbol('Owner');
exports.sybCols = Symbol('Cols');
function convertDefaultValue(toType, val) {
    if (toType === cols.RawType.number)
        return val || 0;
    if (toType === cols.RawType.bool)
        return val || false;
    if (toType === cols.RawType.string)
        return val || '';
    return val;
}
function isContainField(field, filters) {
    if (field === cols.CLogger.updateTime.jsName)
        return true;
    let bool = true;
    for (let i = 0; i < filters.length; i++) {
        const elem = filters[i];
        if (elem === '-')
            bool = false;
        else if (elem === '+')
            bool = true;
        else {
            if (field === elem)
                return bool;
            if (wcl.isFunction(elem)) {
                if (elem(field))
                    return bool;
            }
            else if (elem instanceof cols.Column) {
                if (field === elem.jsName)
                    return bool;
            }
            else if (wcl.isArray(elem)) {
                if (_.some(elem, i => i === field || (i instanceof cols.Column && i.jsName === field)))
                    return bool;
            }
        }
    }
    return !bool;
}
let EntityProxyHandle = {
    getOwner(target) {
        return target[exports.sybOwner];
    },
    get(target, property) {
        let theCol;
        if (property instanceof cols.Column) {
            theCol = property;
            property = property.jsName;
        }
        else if (wcl.isString(property)) {
            let owner = this.getOwner(target);
            if (property === 'hasOwnProperty')
                return Object.prototype.hasOwnProperty.bind(owner.raw);
            theCol = owner.entityType.cols[property];
            if (!theCol) {
                if (property[0] === '$')
                    property = property.substr(1);
                if (property === 'ow')
                    return owner;
                let val = owner[property];
                if (val && wcl.isFunction(val))
                    return val.bind(owner);
                if (property === 'toJSON' || property === 'then')
                    return;
                if (!owner.hasOwnProperty(property) && !owner.__proto__.hasOwnProperty(property))
                    throw new Error(`表'${owner.entityType.tableName}'不存在字段: '${property}(区分大小写)'`);
                return val;
            }
        }
        let val = Reflect.get(target, property);
        if (!theCol || !(theCol instanceof cols.Column))
            return val;
        val = convertDefaultValue(theCol.type.rawType, val);
        if (theCol.type.isFixLength && wcl.isString(val))
            val = val.trimRight();
        return val;
    },
    set(target, property, val) {
        let theCol;
        let owner = this.getOwner(target);
        if (property instanceof cols.Column) {
            theCol = property;
            property = property.jsName;
        }
        else if (wcl.isString(property)) {
            theCol = owner.entityType.cols[property];
            if (!theCol) {
                if (property[0] === '$')
                    property = property.substr(1);
                if (!owner.hasOwnProperty(property) && !owner.__proto__.hasOwnProperty(property))
                    throw new Error(`表'${owner.entityType.tableName}'不存在字段: '${property}(区分大小写)'`);
                Reflect.set(owner, property, val);
                return true;
            }
        }
        if (owner.state === EntityState.normal)
            owner.edit();
        if (owner.state === EntityState.edit && !owner.oldValues.hasOwnProperty(property))
            owner.oldValues[property] = Reflect.get(target, property);
        val = convertDefaultValue(theCol.type.rawType, val);
        return Reflect.set(target, property, val);
    },
};
var EntityState;
(function (EntityState) {
    EntityState["new"] = "";
    EntityState["edit"] = "";
    EntityState["delete"] = "";
    EntityState["normal"] = "";
})(EntityState = exports.EntityState || (exports.EntityState = {}));
wcl.defConsts(EntityState);
function hi(db) {
    return db || dbApi.main;
}
class StaticEntity {
    static get cols() {
        return this[exports.sybCols];
    }
    static get titleCols() {
        return this[exports.sybCols][cols.sybOption].titles;
    }
    static get tableName() {
        let theCols = this.cols;
        let option = cols.getColsOption(theCols);
        return option.tableName;
    }
    static get primaryKeys() {
        let theCols = this.cols;
        let option = cols.getColsOption(theCols);
        return option.keys;
    }
    static getSelect(where) {
        return __awaiter(this, void 0, void 0, function* () {
            return getSelect(this, where);
        });
    }
    static select(where, option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let sql = getSelect(this, where);
            return hi(option.db).qry(sql);
        });
    }
    static one(where, option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let r = yield this.select(where, option);
            return r && r.length > 0 ? this.edit(r[0]) : undefined;
        });
    }
    static id(value, option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let theCols = this.cols;
            let colsOption = cols.getColsOption(theCols);
            let where = {};
            where[colsOption.keys[0]] = value;
            return yield this.one(where, option);
        });
    }
    static key(colOrColumnName, value, option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let theCols = this.cols;
            let colsOption = cols.getColsOption(theCols);
            let where = {};
            where[colOrColumnName] = value;
            return yield this.one(where, option);
        });
    }
    static all(where, option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let rs = yield this.select(where, option);
            let ret = rs && rs.map(i => this.edit(i), this);
            ret && (ret.columns = this.titleCols);
            return ret;
        });
    }
    static getInsert(insertContent) {
        let obj = new this(insertContent, true);
        return obj.getInsert();
    }
    static insert(insertContent, option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let sql = this.getInsert(insertContent);
            return hi(option.db).exe(sql, option.trans);
        });
    }
    static getUpdate(updateContent, where, option = wcl.empty) {
        let sql = getUpdate(this, updateContent, where, option.fields);
        return sql;
    }
    static update(updateContent, where, option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let sql = this.getUpdate(updateContent, where, option);
            return hi(option.db).exe(sql, option.trans);
        });
    }
    static getDelete(where) {
        where = getWhere(this, where);
        if (!where)
            throw new Error('因为delete是个危险的动作，必须添加where条件语句，请检查你的条件语句!');
        return `delete [${this.tableName}] where ${where}`;
    }
    static delete(where, option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let sql = this.getDelete(where);
            return hi(option.db).exe(sql, option.trans);
        });
    }
    static new(rawObj) {
        return new this(rawObj, true).val;
    }
    static edit(rawObj) {
        return new this(rawObj, false).val;
    }
    static trimFixLength(rawOrEntity) {
        if (!rawOrEntity)
            return;
        let cols = this.cols;
        for (const key in rawOrEntity) {
            if (rawOrEntity.hasOwnProperty(key)) {
                if (cols[key] && cols[key].type.isFixLength) {
                    const elem = rawOrEntity[key];
                    if (wcl.isString(elem))
                        rawOrEntity[key] = elem.trimRight();
                }
            }
        }
        return rawOrEntity;
    }
}
class Entity extends StaticEntity {
    constructor(rawObj, isNew) {
        super();
        this.state = EntityState.normal;
        this.touch(rawObj, isNew);
    }
    get entityType() {
        return this.__proto__.constructor;
    }
    touch(rawObj, isNew) {
        if (isNew && rawObj) {
            this.raw = wcl.extends({}, rawObj);
        }
        else if (isNew)
            this.raw = {};
        else if (rawObj)
            this.raw = rawObj;
        else
            throw new Error(`准备使用touch或者edit修改对象时，原始对象rawObj不能为空!`);
        this.raw[exports.sybOwner] = this;
        this.oldValues = undefined;
        this.val = new Proxy(this.raw, EntityProxyHandle);
        this.state = isNew || !rawObj ? EntityState.new : EntityState.normal;
        return this.val;
    }
    drop() {
        if (this.state !== EntityState.normal)
            throw `状态为'${this.state}'的对象，不允许调用drop方法`;
        this.state = EntityState.delete;
    }
    edit() {
        if (this.state !== EntityState.normal)
            throw `状态为'${this.state}'的对象，不允许调用edit方法`;
        this.state = EntityState.edit;
        this.oldValues = {};
    }
    get isChanged() {
        return this.state === EntityState.edit &&
            Boolean(this.oldValues) && Object.keys(this.oldValues).length > 0;
    }
    isModified(col) {
        if (!this.oldValues)
            return false;
        if (col instanceof cols.Column)
            col = col.jsName;
        return this.oldValues.hasOwnProperty(col);
    }
    accept() {
        this.state = EntityState.normal;
        this.oldValues = undefined;
    }
    reject(property) {
        if (this.state !== EntityState.edit)
            return;
        if (property) {
            if (property instanceof cols.Column)
                property = property.jsName;
            if (this.oldValues.hasOwnProperty(property)) {
                Reflect.set(this.raw, property, this.oldValues[property]);
                delete this.oldValues[property];
            }
        }
        else {
            wcl.extends_(this.raw, this.oldValues);
            this.state = EntityState.normal;
            this.oldValues = undefined;
        }
    }
    getPrimaryKeyPart() {
        return getPrimaryKeyPart(this.entityType.primaryKeys, this.raw, true);
    }
    getInsert() {
        let theCols = this.entityType.cols;
        let option = theCols[cols.sybOption];
        if (option.logAble)
            this.val[cols.CLogger.createTime.jsName] = new Date();
        if (option.keys[0].type.baseType === cols.ColType.bigint
            && !this.val[option.keys[0]])
            this.val[option.keys[0]] = wcl.newId;
        let sb = new wcl.stringBuilder();
        sb.append('insert into [', this.entityType.tableName, '] (');
        for (const key in theCols) {
            let elem = theCols[key];
            sb.append('[', elem.fieldName, '], ');
        }
        sb.backChar(2);
        sb.append(') values (');
        let col;
        try {
            for (const key in theCols) {
                col = theCols[key];
                sb.append(col.valuePart(this.val[col.jsName], false), ', ');
            }
        }
        catch (err) {
            throw new Error(`获取字段'${col.jsName}'的值时出现错误\r\n${err.message}`);
        }
        sb.back();
        sb.append(')');
        return sb.toString();
    }
    getUpdate(option = wcl.empty) {
        let onlyModified = option.onlyModified;
        if (onlyModified == undefined)
            onlyModified = true;
        let theCols = this.entityType.cols;
        let colOption = theCols[cols.sybOption];
        if (colOption.logAble)
            this.val[cols.CLogger.updateTime.jsName] = new Date();
        let fields = toCols(theCols, option.fields);
        let sb = new wcl.stringBuilder();
        sb.append('update [', this.entityType.tableName, '] set ');
        let found = false;
        let col;
        try {
            for (const key in theCols) {
                col = theCols[key];
                if (col.primaryKey)
                    continue;
                if (onlyModified && !this.isModified(col))
                    continue;
                if (fields && !isContainField(col.jsName, fields))
                    continue;
                found = true;
                sb.append(col.valuePart(this.val[col.jsName], true), ', ');
            }
        }
        catch (err) {
            throw new Error(`获取字段'${col.jsName}'的值时出现错误\r\n${err.message}`);
        }
        if (!found)
            return '';
        sb.back();
        sb.append(' where ', this.getPrimaryKeyPart());
        return sb.toString();
    }
    getDelete() {
        return `delete [${this.entityType.tableName}] where ${this.getPrimaryKeyPart()}`;
    }
    getActionSql() {
        switch (this.state) {
            case EntityState.new: return this.getInsert();
            case EntityState.edit: return this.getUpdate();
            case EntityState.delete: return this.getDelete();
        }
    }
    insert(option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let sql = this.getInsert();
            let ret = yield hi(option.db).exe(sql, option.trans);
            this.accept();
            return ret;
        });
    }
    update(option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let sql = this.getUpdate(option);
            let ret = yield hi(option.db).exe(sql, option.trans);
            this.accept();
            return ret;
        });
    }
    delete(option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let sql = this.getDelete();
            let ret = yield hi(option.db).exe(sql, option.trans);
            return ret;
        });
    }
    commit(option = wcl.empty) {
        return __awaiter(this, void 0, void 0, function* () {
            let sql = this.getActionSql();
            let ret = yield hi(option.db).exe(sql, option.trans);
            this.accept();
            return ret;
        });
    }
}
exports.Entity = Entity;
function getPrimaryKeyPart(keys, content, mustExists) {
    let r = '';
    for (const col of keys) {
        if (mustExists && !content.hasOwnProperty(col.jsName))
            throw new Error(`条件语句中未指定主键字段'${col.jsName}'的值!`);
        if (r)
            r += ' and ';
        r += col.valuePart(content[col.jsName], true);
    }
    return r;
}
function getWhere(entityClass, where) {
    let theCols = entityClass[exports.sybCols] || entityClass;
    if (wcl.isObject(where) && !(where instanceof wcl.stringBuilder)) {
        let sb = wcl.sb();
        for (const key in where) {
            let theCol = theCols[key];
            if (!theCol)
                continue;
            sb.append('[', theCol.fieldName, '] = ', theCol.valuePart(where[key]), ' and ');
        }
        sb.back();
        where = sb;
    }
    return where && where.toString();
}
function getSelect(entityClass, where) {
    let theCols = entityClass[exports.sybCols] || entityClass;
    let option = cols.getColsOption(theCols);
    let all = option.selectAll;
    if (!all) {
        let sb = new wcl.stringBuilder();
        sb.append('select ');
        for (const key in theCols) {
            const elem = theCols[key];
            if (elem.fieldName === elem.jsName)
                sb.append('[', elem.fieldName, '], ');
            else
                sb.append('[', elem.fieldName, '] as ', elem.jsName, ', ');
        }
        sb.backChar(2);
        sb.append(' from [', option.tableName, ']');
        all = option.selectAll = sb.toString();
    }
    where = getWhere(entityClass, where);
    if (!where)
        return all;
    return all + ' where ' + where;
}
function getUpdate(entityClass, updateContent, where, fields) {
    let theCols = entityClass[exports.sybCols] || entityClass;
    let option = cols.getColsOption(theCols);
    if (option.logAble)
        updateContent[cols.CLogger.updateTime.jsName] = new Date();
    fields = toCols(theCols, fields);
    let sb = new wcl.stringBuilder();
    sb.append('update [', option.tableName, '] set ');
    let found = false;
    let col;
    try {
        for (const key in updateContent) {
            col = theCols[key];
            if (!col || col.primaryKey)
                continue;
            if (fields && !isContainField(col.jsName, fields))
                continue;
            found = true;
            sb.append(col.valuePart(updateContent[key], true), ', ');
        }
    }
    catch (err) {
        throw new Error(`获取字段'${col.jsName}'的值时出现错误\r\n${err.message}`);
    }
    if (!found)
        return '';
    sb.back();
    sb.append(' where ');
    if (where && (where = getWhere(entityClass, where)))
        sb.append(where);
    else
        sb.append(getPrimaryKeyPart(option.keys, updateContent, true));
    return sb.toString();
}
function toCols(theCols, fields) {
    if (!fields)
        return;
    for (let i = 0; i < fields.length; i++) {
        const elem = fields[i];
        if (wcl.isString(elem) && elem !== '-' && elem !== '+')
            fields[i] = theCols[elem];
    }
    return fields;
}
function toFields(fields) {
    if (!fields)
        return;
    for (let i = 0; i < fields.length; i++) {
        const elem = fields[i];
        if (elem instanceof cols.Column)
            fields[i] = elem.jsName;
    }
    return fields;
}
//# sourceMappingURL=entity.js.map