"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
function __export(m) {
    for (var p in m) if (!exports.hasOwnProperty(p)) exports[p] = m[p];
}
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const wcl = require("../../jswcl-ts");
const _ = require("lodash");
const cols = __importStar(require("./column"));
const dbApi = __importStar(require("./db"));
exports.db = dbApi;
const ent = __importStar(require("./entity"));
function getAllFields(tableName) {
    return `SELECT
            t.name AS 'TableName',
            c.name AS 'FieldName',
            c.colid As 'FieldPos',
            p.name AS UserType,
            p2.name AS Type,
            c.length,
            c.xprec,
            c.xscale,
            c.isnullable
        FROM
            sysobjects t
            JOIN Syscolumns c ON t.id = c.id
            JOIN systypes p ON c.xusertype = p.xusertype
            JOIN systypes p2 ON p.xtype = p2.xtype AND p.xtype = p2.xusertype
        WHERE
            t.xtype = 'U' AND t.name ='${tableName}'
        ORDER BY
            t.name, c.name`;
}
function nolog(sql) {
    return new wcl.cfgableString(sql);
}
function getCreateTable(entityClass, option = wcl.empty) {
    let theCols = entityClass[ent.sybCols] || entityClass;
    let colOption = cols.getColsOption(theCols);
    let sb = wcl.sb();
    if (option.dropFirst) {
        sb.appendLine(`IF OBJECT_ID ('dbo.${colOption.tableName}') IS NOT NULL
    DROP TABLE dbo.[${colOption.tableName}];`);
    }
    else if (option.ifNotExists) {
        sb.appendLine(`IF OBJECT_ID ('dbo.${colOption.tableName}') IS NULL`);
    }
    sb.appendLine('create table [', colOption.tableName, '] (');
    for (const key in theCols) {
        const col = theCols[key];
        sb.append(`   [${col.fieldName}] \t\t ${col.toSqlType()}`);
        if (col.notnull)
            sb.append(' not null,\r\n');
        else
            sb.append(',\r\n');
    }
    sb.append(`   constraint PK_${colOption.tableName} primary key (`);
    for (const key in colOption.keys) {
        const col = colOption.keys[key];
        sb.append(col.fieldName, ', ');
    }
    sb.back();
    sb.appendLine(')');
    sb.appendLine(')');
    let r = [];
    r.push(sb.toString());
    sb.clear();
    let indexes = theCols[cols.sybIndex];
    if (indexes) {
        for (const key in indexes) {
            let fields = indexes[key].left(':');
            let opt = indexes[key].right(':').trim(true, false);
            let where;
            if (opt !== 'uq')
                where = opt;
            sb.appendLine(`CREATE ${opt ? 'UNIQUE' : ''} INDEX ${key} ON dbo.[${colOption.tableName}] (${fields}) ${where};`);
        }
    }
    r.push(sb.toString());
    return r;
}
exports.getCreateTable = getCreateTable;
function upgrade(entityClass, db, option = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        let theCols = entityClass[ent.sybCols] || entityClass;
        let colOption = cols.getColsOption(theCols);
        let sql;
        if (option.forceRecreate) {
            let sqls = getCreateTable(theCols, { dropFirst: true });
            for (const i of sqls) {
                yield db.exe(nolog(i));
            }
            return 1;
        }
        else {
            sql = `SELECT CASE WHEN OBJECT_ID ('dbo.${colOption.tableName}') > 0 THEN 1 ELSE 0 END`;
            let exists = yield db.val(nolog(sql));
            let count = 0;
            if (exists)
                count = yield db.val(nolog(`select count(1) from [${colOption.tableName}]`));
            if (!exists || count === 0) {
                let sqls = getCreateTable(theCols, { dropFirst: exists });
                for (const i of sqls) {
                    yield db.exe(nolog(i));
                }
                return 1;
            }
        }
        sql = getAllFields(colOption.tableName);
        let rs = yield db.qry(nolog(sql));
        let updateFields = '';
        for (const key in theCols) {
            const elem = theCols[key];
            let field = _.find(rs, i => i.FieldName === elem.fieldName);
            if (!field) {
                let notNull = '';
                if (elem.notnull)
                    notNull = elem.type.needQuoted ? "not null default ''" : "not null default 0";
                let add = `alter table [${colOption.tableName}] add [${elem.fieldName}] ${elem.toSqlType()} ${notNull}`;
                updateFields += add + '\r\n';
                continue;
            }
            let newType = elem.type.baseType;
            if (newType.name !== field.Type) {
                log(`无法自动升级表'${colOption.tableName}'的字段'${elem.fieldName}'类型: ${field.Type}->${newType.name}`.warn);
                continue;
            }
            newType = elem.type;
            if (newType.hasLengthProp) {
                let isFloat = newType.prec !== 0 || field.xscale !== 0;
                if (isFloat ? newType.length !== field.xprec || newType.prec !== field.xscale : newType.length !== 0 && newType.length !== field.length) {
                    let notNull = '';
                    if (elem.notnull)
                        notNull = elem.type.needQuoted ? "not null default ''" : "not null default 0";
                    let alter = `alter table [${colOption.tableName}] alter column [${elem.fieldName}] ${elem.toSqlType()} ${notNull}`;
                    updateFields += alter + '\r\n';
                    continue;
                }
            }
        }
        _.each(rs, field => {
            if (!_.some(theCols, i => i.fieldName === field.FieldName))
                log(`无法自动升级表'${colOption.tableName}'，删除字段'${field.FieldName}'`.warn);
        });
        if (updateFields) {
            log('auto update fields...');
            log(updateFields);
            yield db.exe(nolog(updateFields));
        }
    });
}
exports.upgrade = upgrade;
let classList = {};
function reg(entityClass) {
    classList[entityClass.tableName] = entityClass;
}
exports.reg = reg;
function get(entityClassName, errIfNotfount = true) {
    let ret = classList[entityClassName];
    if (!ret && errIfNotfount)
        throw new Error(`没有发现已注册的实体类型名称: ${entityClassName}`);
    return ret;
}
exports.get = get;
function upgradeAll(db, option = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        if (dbApi.main.hostId === dbApi.struct.hostId && option.forceRecreate && !option.isWush)
            throw new Error(`为了防止误操作, 不允许在主库和结构库是同一个库(${dbApi.main.hostId})的时候执行强制重建的工作!`);
        db = db || dbApi.struct;
        for (const i in classList) {
            yield upgrade(classList[i], db, option);
        }
    });
}
exports.upgradeAll = upgradeAll;
function autoUpgradeAll(db, option = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        if (db && !db.hostId) {
            option = db;
            db = undefined;
        }
        if (option.autoWithStructDb) {
            let tmpOption = wcl.extends({}, option);
            tmpOption.forceRecreate = true;
            yield upgradeAll(dbApi.struct, tmpOption);
        }
        if (!dbApi.isAutoUpgrade)
            return;
        db = db || dbApi.main;
        if (db !== dbApi.struct || !option.autoWithStructDb)
            yield upgradeAll(db, option);
    });
}
exports.autoUpgradeAll = autoUpgradeAll;
__export(require("./column"));
__export(require("./entity"));
function make(tableName, theCols, entityType, option = wcl.empty) {
    let colOption = theCols[cols.sybOption] || {};
    colOption.logAble = option.logAble;
    colOption.tableName = tableName;
    theCols[cols.sybOption] = colOption;
    cols.makeColumns(theCols);
    entityType[ent.sybCols] = theCols;
    if (option.reg == undefined || option.reg)
        exports.reg(entityType);
    _.each(entityType, (i, key) => {
        if (wcl.isArray(i) && i[0] === '@grp') {
            i.shift();
            entityType[key] = cols.getGroupFields(theCols, i);
        }
    });
}
exports.make = make;
//# sourceMappingURL=index.js.map