const wcl = require('./wcl');
let cst = wcl.cst;

const TAB = wcl.char.tab;
const RETURN = wcl.char.return;
const NEWLINE = wcl.char.newline;
const ESCAPE = wcl.char.escape;
const DOUBLE_QUOTED = wcl.char.double_quoted;

/**
 * @param {String} text 
 * */
function __escape(text) {
    let pos = 0;
    let length = text.length;

    //多数字符串不需要转义，所以这里先优化处理（先判断不需要转义的字符串）
    while (true) {
        if (pos === length)
            return text;

        let ch = text.charCodeAt(pos);
        if (ch === DOUBLE_QUOTED || ch === TAB || ch === RETURN || ch === NEWLINE || ch === ESCAPE) //在第一次出现转义字符进入转义处理。
            break;
        pos++;
    }

    let saveIndex = 0;
    let sb = new wcl.stringBuilder();
    while (pos < length) {
        let ch = text.charCodeAt(pos);
        if (ch === DOUBLE_QUOTED)
            ch = '\\"';
        else if (ch === TAB)
            ch = '\\t';
        else if (ch === RETURN)
            ch = '\\r';
        else if (ch === NEWLINE)
            ch = '\\n';
        else if (ch === ESCAPE)
            ch = '\\\\';
        else
            ch = '';
        if (ch) {
            if (pos > saveIndex)
                sb.appendOne(text.slice(saveIndex, pos));
            sb.appendOne(ch);
            saveIndex = ++pos;
        }
        else
            pos++;
    }
    if (pos > saveIndex)
        sb.appendOne(text.slice(saveIndex, pos));
    return sb.toString();
}

function toJson(obj, { fmt, arrElemNewLine, escape } = toJson.option) {
    let sb = wcl.sb();
    let deep = 0;
    let tab = '';
    function inc(count) {
        if (!fmt)
            return;
        deep += count | 1;
        tab = ' '.repeat(deep * 4);
    }
    function line() {
        sb.append.apply(sb, arguments);
        if (fmt)
            sb.appendOne('\n');
    }
    function append(obj) {
        let type = typeof obj;
        if (type === cst.num)
            sb.append(obj);
        else if (type === cst.bool)
            sb.append(obj ? 'true' : 'false');
        else if (type === cst.str)
            sb.append('"', escape ? __escape(obj) : obj, '"');
        else if (type === cst.undef)
            sb.append('null');
        else if (obj instanceof Date) {
            sb.append('"', obj, '"');
        }
        else if (obj === null)
            sb.append('null');
        else if (wcl.isObject(obj)) {
            if (list.indexOf(obj) >= 0)//检测循环应用
                return;
            list.push(obj);

            line('{');
            inc();
            for (const key in obj) {
                const val = obj[key];
                if (wcl.isFunction(val))
                    continue;
                if (fmt)
                    sb.append(tab, '"', key, '": ');
                else
                    sb.append('"', key, '":');
                append(val);
                line(',');
            }
            sb.back(fmt ? 2 : 1);
            line();
            inc(-1);
            sb.append(tab, '}');
        }
        else if (wcl.isArray(obj)) {
            if (obj.length === 0) {
                sb.appendOne('[]');
                return;
            }
            if (arrElemNewLine)
                line('[');
            else
                sb.appendOne('[');
            inc();
            obj.forEach(i => {
                if (arrElemNewLine)
                    sb.appendOne(tab);
                append(i);
                if (arrElemNewLine)
                    line(',');
                else if (fmt)
                    sb.appendOne(', ');
                else
                    sb.appendOne(',');
            });
            sb.back(fmt && arrElemNewLine ? 2 : 1);
            if (arrElemNewLine)
                line();
            inc(-1);
            if (arrElemNewLine)
                sb.append(tab, ']');
            else
                sb.appendOne(']');
        }
    }

    let type = typeof obj;
    if (type === cst.bool)
        sb.append(obj ? 'true' : 'false');
    else if (type === cst.num)
        return obj.toString();
    else if (type === cst.str)
        return obj;
    else if (type === cst.undef)
        return 'null';
    else if (obj instanceof Date)
        return obj.toLocaleString();
    else if (obj === null)
        return 'null';

    let list = [];
    append(obj);

    return sb.toString();
}
toJson.option = { fmt: true, arrElemNewLine: false, escape: true };

toJson.test = function (option) {
    let data = require('../test/testToJson-data1');
    data.xday = new Date();
    data.ttt = undefined;

    let json = this(data);
    data = JSON.parse(json);
    option || (option = toJson.option);
    let a = JSON.stringify(data, undefined, option.fmt ? 4 : 0);
    let b = wcl.toJson(data, wcl.extends({}, [option, { arrElemNewLine: true }]));
    let c = wcl.toJson(data, option);
    log(a === b ? '正确，与JSON.stringify的往返一致' : '错误，与JSON的往返不一致');
    if (a === b)
        log(c);
    else {
        log(a, '\n');
        log(b, '\n');
        log(wcl.truncSameText(a, b, { returnString: true }));
    }
}

module.exports = toJson;