const fs = require('fs');
const path = require('path');

// let replace1 = /\\(2,)/g;
let replace2 = /\\/g;
let empty = new Proxy({}, {
    get(t, p) {
        return undefined;
    },
    set(t, p, v) {
        throw new Error(`'empty'对象不能修改(->empty.${p})!`);
    }
});
let isHandleAppErr = 1;//是否已加载默认错误处理。初始化设置为1，将导致不加载默认错误处理器。

let abortErr = new Error('执行流程已中断');
abortErr.isAbort = true;

/**@param {...string[]} subPath */
function createFullPath__(subPath) {
    function existsAt(endIndex, callback) {
        let thePath = subPath.substr(0, endIndex);
        fs.exists(thePath, bool => {
            if (bool)
                callback(endIndex);
            else {
                endIndex = subPath.lastIndexOf(path.sep, endIndex - 1);
                existsAt(endIndex, callback);
            }
        })
    }
    function make(endIndex, callback) {
        if (endIndex >= subPath.length - 1) {
            callback();
            return;
        }
        endIndex = subPath.indexOf(path.sep, endIndex + 1);
        if (endIndex < 0)
            endIndex = subPath.length;
        let thePath = subPath.substr(0, endIndex);

        fs.mkdir(thePath, err => {
            log('create', thePath);
            if (err)
                callback(err);
            else
                make(endIndex, callback);
        });
    }

    return new Promise((resolve, reject) => {
        existsAt(subPath.length - 1, index => {
            if (index === subPath.length - 1)
                resolve(subPath);
            else {
                make(index, err => {
                    if (err)
                        reject(err);
                    else
                        resolve(subPath);
                })
            }
        });
    });
}

class WCL {
    constructor() {
        this.cst = require('./knowConst');
        this.option = require('./option');
        this.wcl = this;
        this.lastLogTime = undefined;
        this.empty = empty;
    }

    abort() {
        throw abortErr;
    }
    isAbort(err) {
        return err == abortErr || err.isAbort;
    }

    get line() {
        return this.option.line;
    }

    get logLine() {
        return this.option.logLine || '-'.repeat(20);
    }

    pause(msg) {
        this.readline(text => process.exit(),
            msg || '按回车键结束: ');
    }

    readline(callback, msg) {
        const readline = require('readline');

        const read = readline.createInterface({
            input: process.stdin,
            output: process.stdout
        });

        read.question(msg || '请输入一个值(后按回车): ', callback);
    }

    /**@param {...string[]} subPath */
    createFullPath(subPath) {
        subPath = path.join.apply(path, [...arguments, '/']);
        return createFullPath__(subPath);
    }

    appConfigPath(subPath, autoCreate) {
        subPath = path.join(process.cwd(), 'AppConfig', subPath, '/');
        if (autoCreate == undefined || autoCreate)
            return createFullPath__(subPath);
        else
            return subPath;
    }

    appDataPath(subPath, autoCreate) {
        subPath = path.join(process.cwd(), 'AppData', subPath, '/');
        if (autoCreate == undefined || autoCreate)
            return createFullPath__(subPath);
        else
            return subPath;
    }

    appTmpPath(subPath, autoCreate) {
        subPath = path.join(process.cwd(), 'Temp', subPath, '/');
        if (autoCreate == undefined || autoCreate)
            return this.createFullPath__(subPath);
        else
            return subPath;
    }

    //根据需要，自动创建带上下文调用的函数
    fn(fnTarget, context) {
        // return fnTarget && (context ? function () { fnTarget.apply(context, arguments); } : fnTarget);
        if (!fnTarget)
            return;
        if (context)
            return fnTarget.bind(context);
        return fnTarget;
    }

    //是否是对象
    isObject(obj) {
        return Object.prototype.toString.call(obj) === '[object Object]';
    }

    //是否是数组
    isArray(obj) {
        return Object.prototype.toString.call(obj) === '[object Array]';
    }

    //是否是函数
    isFunction(obj) {
        return typeof obj === this.cst.fun;
    }

    //是否是异步函数
    isAsyncFunction(obj) {
        return this.isFunction(obj) && obj.constructor.name === 'AsyncFunction';
    }

    //是否是字符串类型
    isString(obj) {
        return typeof obj === this.cst.str;
    }

    //是否是数值类型
    isNumber(obj) {
        return typeof obj === this.cst.num;
    }

    //返回所创建的新对象数组
    makeObjArray(objDefine, valueArrayArray) {
        if (arguments.length === 1 && !this.isArray(objDefine)) { //一个参数使用{names:xxxx,values:[[xxxx],[xxxxx]]}这种对象。
            valueArrayArray = objDefine.values;
            objDefine = objDefine.names;
        }
        if (arguments.length === 1) {
            valueArrayArray = objDefine;
            objDefine = null;
        }
        var r = {};
        if (!objDefine || this.isArray(objDefine[0]))
            objDefine = ['id', 'val'];
        for (var k = 0; k < valueArrayArray.length; k++) {
            var n = {};
            for (var i = 0; i < objDefine.length; i++) {
                n[objDefine[i]] = valueArrayArray[k][i];
            }
            r[valueArrayArray[k][0]] = n;
        }
        return r;
    }

    // 将mergeFrom拥有的属性合并复制和覆盖到mergeTo里。
    mergeObj(mergeFrom, mergeTo, onlyExists) {
        throw new Error('wcl.mergeObj不在受支持，请改用wcl.extends');
        for (var i in mergeFrom) {
            if (!onlyExists || mergeTo.hasOwnProperty(i)) {
                mergeTo[i] = mergeFrom[i];
            }
        }
        return mergeTo;
    }

    /**
     * 将一列列对象的属性扩展到目标对象，并返回目标对象。    
     * @param {*} toObj 
     * @param {*} fromObj 
     * @param {{onlyExists?:boolean, ignoreExists?: boolean} }
     */
    extends(toObj, fromObj, { onlyExists, ignoreExists } = empty) {
        if (!toObj)
            toObj = {};
        if (fromObj == undefined)
            return toObj;

        let wcl = this;
        function doit(toObj, fromObj) {
            for (let name in fromObj) {
                if (!Object.hasOwnProperty.call(fromObj, name))
                    continue;
                let exists = Object.hasOwnProperty.call(toObj, name);
                if (!exists && onlyExists)
                    continue;
                if (exists && ignoreExists && !wcl.isObject(toObj[name]))
                    continue;
                if (!exists)
                    toObj[name] = fromObj[name];
                else {
                    let toValue = toObj[name];
                    let fromValue = fromObj[name];
                    if (wcl.isObject(toValue)) {
                        if (wcl.isObject(fromValue))
                            doit(toValue, fromValue);
                    }
                    else if (wcl.isArray(toValue)) {
                        if (wcl.isArray(fromValue))
                            toObj[name] = toValue.concat(fromValue);
                        else
                            toValue.push(fromValue);
                    }
                    else
                        toObj[name] = fromValue;
                }
            };
        }

        if (wcl.isArray(fromObj)) {
            fromObj.forEach(i => {
                doit(toObj, i);
            });
        }
        else
            doit(toObj, fromObj);

        return toObj;
    }

    //转化为 prop1=value1;prop2=value2;...的形式
    nvls(obj, spea) {
        if (obj) {
            if (this.isString(obj))
                return this.valueInfo(obj);
            var s = "";
            // spea || (spea = '<br/>');
            spea || (spea = this.line);
            for (var i in obj) {
                s += i + '=' + this.valueInfo(obj[i]) + spea;
            }
            return s;
        }
    }

    //返回对象的属性名称描述字符串
    metaInfo(obj, withValue) {
        var s = "";
        var k = 0;
        for (var i in obj) {
            if (k > 0)
                s += ', ';
            s += i;
            if (withValue)
                s += ' = ' + this.valueInfo(obj[i]);
            k++;
            if (k === 7) {
                s += '\r\n';
                k = 0;
            }
        }
        return s;
    }

    //是否是空白字符串
    isWhite(text) {
        return !text || !text.toString().trim();
    }

    //是null时返回空字符串，否则返回text本身
    notNull(text) {
        return text || '';
    }

    //根据是否为空，自动连接字符串
    merge(left, concat, right) {
        if (!right)
            return left;
        else if (!left)
            return right;
        return left + concat + right;
    }

    //格式化数字
    fmtNum(num, precision, separator) {
        var parts;
        // 判断是否为数字
        if (!isNaN(parseFloat(num)) && isFinite(num)) {
            // 把类似 .5, 5. 之类的数据转化成0.5, 5, 为数据精度处理做准, 至于为什么
            // 不在判断中直接写 if (!isNaN(num = parseFloat(num)) && isFinite(num))
            // 是因为parseFloat有一个奇怪的精度问题, 比如 parseFloat(12312312.1234567119)
            // 的值变成了 12312312.123456713
            num = Number(num);
            // 处理小数点位数
            num = (typeof precision !== this.cst.undef ? num.toFixed(precision) : num).toString();
            // 分离数字的小数部分和整数部分
            parts = num.split('.');
            // 整数部分加[separator]分隔, 借用一个著名的正则表达式
            parts[0] = parts[0].toString().replace(/(\d)(?=(\d{3})+(?!\d))/g, '$1' + (separator || ','));

            return parts.join('.');
        }
        return 0;
    }

    //格式化一个值
    fmtValue(value, format) {
        if (!format)
            return this.prnOf(value);
        format = format.split(',');
        var type = format[0];
        if (type === 'n')
            return format.length > 1 ? this.fmtNum(value, Number(format[1] || '0'), format[2]) : value || '0';
        else if (type === 'd')
            return new Date(value).fmt(format[1] || 'D');
        else if (type === 'b')
            return value ? '√' : '';
        return value;
    }

    //获取值的提示信息。func, obj, 或者toString()值。
    valueInfo(obj) {
        if (this.isFunction(obj))
            return '[func]';
        if (obj === '[object Object]')
            return '[obj]';
        if (obj === false)
            return 'false';
        if (obj === true)
            return 'true';
        if (obj === 0)
            return '0';
        if (!obj)
            return this.cst.undef;
        let isString = typeof (obj) == this.cst.str;
        obj = obj.toString();
        obj = obj.replace('\r', '\\r').replace('\n', '\\n').replace('\t', '\\t');
        if (isString)
            return obj.qq;
        return obj;
    }

    //获取对象的显示值。（为null,undefine时转换成空字符串）
    prnOf(obj) {
        if (obj === 0)
            return 0;
        else if (obj === false)
            return false;
        else
            return obj || '';
    }

    //通过随机数生成一个新的guid
    get guid() {
        return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
            var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
            return v.toString(16);
        });
    }

    get newId() {
        return this.idHelper.newId;
    }

    defConsts(sConst) {
        if (typeof (sConst) === this.cst.str) {
            var a = sConst.split(',');
            var r = {};
            for (var i in a) {
                var p = a[i].indexOf('=');
                if (p < 1)
                    r[a[i]] = a[i];
                else
                    r[a[i].substr(0, p)] = a[i].substr(p + 1, a[i].length - p - 1);
            }
            return r;
        }
        else {
            for (const key in sConst) {
                sConst[key] = key;
            }
        }
    }

    //删除json中的 /*....*/ 注释部分
    removeComment(srcJson) {
        return srcJson && srcJson.replace(/\/\*.*?\*\//g, '');
    }

    //将1段手写的json代码（最外面不包含{}）计算出结果后返回。
    getData(text) {
        eval('var r ={' + text + '}');//r的范围在本函数内
        return r;
    }

    //休眠指定的毫秒数（js为单线程）
    sleep(millSeconds) {
        var now = new Date();
        var exitTime = now.getTime() + millSeconds;
        while (true) {
            now = new Date();
            if (now.getTime() > exitTime)
                return;
        }
    }

    setLog() {
        if (arguments.length === 0)
            return console.originalLog !== console.log;
        if (arguments.length === 1) {
            let a = arguments[0];
            if (a === 0)
                global.log = console.originalLog;
            else if (a === 1)
                global.log = this.log.bind(this);
            console.log = global.log;
        }
    }

    log() {
        // 【2018-03-31 07:11 by wush】
        // if (option.logTime) {
        //     process.stdout.write(wcl.now('hh:mm:ss').yellow);
        //     process.stdout.write('> '.yellow);
        // }
        /*            
        process.stdout.write在标准cmd控制台中有效，在vscode中无效
        console的相关输出命令都会自动换行
        所以改成下面的方法，重整 arguments 参数。
         */

        /* 由于vscode调试时会将console输出时的堆栈显示出来，
        * 如果调用wcl.log将会导致堆栈显示不够准确（全部会在wcl相应行上） 
        * 为了方便调试和跟踪堆栈信息，这里只能调用原生的console.log方法
        */
        if (arguments.length === 0)
            return;
        if (this.option.logTime && arguments[0] !== this.option.logLine) {
            let useTime = '';
            let lastTime = this.lastLogTime;
            let now = new Date();
            if (this.option.logUseTime) {
                useTime = 0;
                if (lastTime && now > lastTime)
                    useTime = (now - lastTime);
                this.lastLogTime = now;
                if (useTime < 1000)
                    useTime = `(${useTime}ms)`;
                else if (useTime < 60000)
                    useTime = `(${(useTime / 1000).toFixed(2)}s)`;
                else
                    useTime = `(${(useTime / 60000).toFixed(2)}m)`;
            }
            var time = `${now.fmt(this.option.logTimeFormat)}${useTime}> `;

            time = time.prompt;
            arguments[0] = time + (arguments[0] || '(null)').toString();
        }
        console.originalLog.apply(console, arguments);
    }

    //返回当前时间。D: yyyy-MM-dd, T: hh:mm:ss, "L": yyyy-MM-dd hh:mm:ss, "D-": yyyyMMdd, "T-": hhmmss, "L-": yyyyMMdd hhmmss
    now(format) {
        return format ? new Date().fmt(format) : new Date();
    }

    //使用 val(obj, "name1.name2.name3...")方式安全的访问对象的属性。确保不会因为undeinfe而报错
    //此方法可以单独拷贝到任何地方，不依赖其他函数。
    val(obj, name) {
        if (!obj) return obj;
        var idx = name.indexOf('.');
        if (idx < 0) return obj[name];
        obj = obj[name.substr(0, idx)];
        return val(obj, name.substr(idx + 1))
    }

    //以递归方式，获取一个对象的子属性名称列表
    getDeepObjLayout(obj, maxDeep) {
        maxDeep || (maxDeep = 5);
        let sb = new this.stringBuilder();
        function get(subObj, deep, tab) {
            if (!subObj)
                return;
            for (let key in subObj) {
                let elem;
                try {
                    elem = subObj[key];
                    sb.appendLine(tab + key);
                } catch (err) {
                    sb.appendLine(tab + key + '(error)');
                }
                if (deep < maxDeep && typeof (elem) === this.cst.obj && !this.isArray(elem))
                    get(elem, deep + 1, tab + '.' + key);
            }
        }
        get(obj, 1, '');
        return sb.toString();
    }

    /**
     * 获取第一个出现的路径扩展名。格式： [extName, start, end]
     * @param  {String} path 目标路径字符串
     * @param {Boolean} fullExtStyle 为true时，这扩展名里允许多个点号(最多只取2个),比如: .d.ts
     * @returns {[]|undefined}
     */
    getFirstExt(path, fullExtStyle) {
        /**@type {Number} */
        let start = path.indexOf('.');
        if (start < 0)
            return ['', -1, -1];
        for (let i = 0; i < start; i++) {
            const char = path[i];
            if (char === '?' || char === '#')
                return ['', -1, -1];
        }

        let end;
        if (fullExtStyle)
            fullExtStyle = 1;
        for (end = start + 1; end < path.length; end++) {
            const char = path[end];
            if (char === '.') {
                if (!start)
                    start = end;
                if (fullExtStyle)
                    fullExtStyle++;
                else
                    start = end;
            }
            if (char === ' ' || char === '\\' || char === '/' || char === '?' || char === '#')
                break;
        }
        if (fullExtStyle && fullExtStyle > 2)
            start = path.lastIndexOf('.', path.lastIndexOf('.', end) - 1);
        return [path.slice(start, end), start, end];
    }

    /**
     * 获取路径中的第一段目录
     */
    getFirstPath(path) {
        if (!path || path.length < 3)
            return path;
        for (let i = 2; i < path.length; i++) {
            const char = path[i];
            if (char === '\\' || char === '/')
                return path.substr(0, i);
            if (char === '?' || char === '#')
                break;
        }
        return '/';
    }

    //开启一个无限等待的循环, 直到call()为true或者到达了指定的最大尝试次数
    wait(call, { ticks = 100, initTicks = 0, maxTimes = 0 } = {}) {
        let times = 0;
        let retry = () => {
            if (call && call(++times))
                return;
            if (--maxTimes !== 0)
                setTimeout(retry, ticks);
        };
        initTicks ? setTimeout(retry, initTicks) : retry();
    }

    joinPath(args) {
        for (let index = 0; index < arguments.length; index++) {
            const elem = arguments[index];
            if (elem === undefined || elem === null)
                arguments[index] = '';
        }
        /**@type {String} */
        let r = path.join.apply(path, arguments);
        return r.replace(replace2, '/');
    }

    getDiffIndex(byText, destText, filePathMode = false) {
        function get() {
            let i = 0;
            while (true) {
                if (i >= destText.length || i >= byText.length)
                    return destText.length === byText.length ? -1 : i;
                if (destText.charCodeAt(i) != byText.charCodeAt(i))
                    return i;
                i++;
            }
        }
        let idx = get();
        if (filePathMode) {
            while (idx > 0) {
                if (destText[idx] === '/' || destText[idx] === '\\')
                    break;
                idx--;
            }
        }
        return idx;
    }

    /**
     * 从destText剔除与byText前导相同的字符部分
     * @param  {String} byText
     * @param  {String} destText
     * @param  {Boolean | undefined} {string
     * @param  {String | undefined} map
     * @param  {Boolean | undefined} filePathMode}
     */
    truncSameText(byText, destText, { returnString, fillWith, filePathMode }) {
        if (!destText)
            return '';
        let idx = this.getDiffIndex(byText, destText, filePathMode);
        if (returnString) {
            if (idx < 0)
                return '';
            if (idx >= destText.length)
                return '';
            if (fillWith && idx > 0)
                return fillWith.repeat(idx) + destText.substr(idx);
            return destText.substr(idx) || '';
        }
        return idx;
    }

    /**@param {string[]} list*/
    formatDirTree(list, sort) {
        if (!list || list.length < 2)
            return list;
        if (sort == undefined || sort)
            list.sort();

        let allSame = '-';
        for (let i = 0; i < list.length - 1; i++) {
            let same = list[i].slice(0, wcl.getDiffIndex(list[i], list[i + 1]));
            if (allSame === '-') {
                allSame = same;
                continue;
            }
            allSame = allSame.slice(0, wcl.getDiffIndex(allSame, same));
            if (!allSame)
                break;
        }

        let last = '';
        for (let i = 0; i < list.length; i++) {
            let elem = list[i];
            list[i] = wcl.truncSameText(last, elem, { returnString: true, fillWith: ' ', filePathMode: true });
            last = elem;
            list[i] = list[i].substr(allSame.length);
        }
        list.unshift(allSame);

        return list;
    }

    /**
     * 将一些类的实现混入到目标类（实现多继承）
     * @param {function} toClass 
     * @param {function[]} fromCalsses 
     */
    mixins(toClass, fromCalsses) {
        fromCalsses.forEach(i => {
            Object.getOwnPropertyNames(i.prototype).forEach(name => {
                toClass.prototype[name] = i.prototype[name];
            })
        });
    }

    handleAppErr() {
        if (isHandleAppErr)
            return;
        function onErr(err, promiseObj) {
            log(`${err.stack.stack}`);
            // process.exit(-1);
        }
        process.on('uncaughtException', onErr);
        process.on('unhandledRejection', onErr);
        isHandleAppErr = true;
    }

    /********************************node js*********************************** */
    reqPath(moduleName, from) {
        if (from)
            return require.resolve(moduleName, from.paths);
        return require.resolve(moduleName);
    }

    readSyn() {
        //参考： https://blog.csdn.net/kiwi_coder/article/details/8623747
        process.stdin.pause();
        var response = fs.readSync(process.stdin.fd, 1000, 0, "utf8");
        process.stdin.resume();
        return response[0].trim();
    }
}

let wcl = new WCL();
WCL.prototype.extends_ = WCL.prototype.extends;
Number.prototype.fmt = function (precision, separator) {
    return wcl.fmtNum(this, precision, separator);
}

module.exports = wcl;

const stringBuilder = require('./stringBuilder');
WCL.prototype.stringBuilder = stringBuilder;
/**@returns {stringBuilder} */
WCL.prototype.sb = function (fldObj, opr, val) {
    let sb = new this.stringBuilder();
    if (arguments.length === 0)
        return sb;
    if (arguments.length === 1)
        return sb.add(fldObj);
    if (arguments.length > 3)
        return sb.add.apply(sb, arguments);
    return sb.fld.apply(sb, arguments);
};
/**@returns {stringBuilder} */
WCL.prototype.fld = function (fldObj, opr, val) {
    let sb = new this.stringBuilder();
    return sb.fld.apply(sb, arguments);
};

WCL.prototype.fromCompactArray = require('./compactArray');
WCL.prototype.testTime = require('./testTime');
WCL.prototype.char = require('./char');
WCL.prototype.script = require('./script');
WCL.prototype.cfgableString = require('./cfgableString');
WCL.prototype.idHelper = require('./idHelper');
WCL.prototype.toJson = require('./toJson');