import _ = require('lodash');

declare namespace wcl {
    const option: {
        line: string,
        attachLog: boolean,
        logTime: boolean,
        logTimeFormat: string,
        logUseTime: boolean,
        logLine: string
    };

    const cst: {
        undef: 'undefined',
        obj: 'object',
        bool: 'boolean',
        str: 'string',
        num: 'number',
        fun: 'function',

        post: 'POST',
        get: 'GET',

        contentType: 'content-type',
        octet_stream: 'application/octet-stream',
        appJson: 'application/json',
        appJson_utf8: 'application/json; charset=utf-8',
        appForm: 'application/x-www-form-urlencoded',
        appForm_utf8: 'application/x-www-form-urlencoded; charset=utf-8',
        multipart: 'multipart/form-data',

        textPlain: 'text/plain',
        textPlain_utf8: 'text/plain; charset=utf-8',
        textHtml: 'text/html',
        textHtml_utf8: 'text/html; charset=utf-8',

        utf8: 'utf-8',

        isAjax: 'isajax',

        apiErr: {
            unknow: 1000,
            normal: 1001,
            parameter: 1002,
            unlogin: 1003,
        }
    }
}

declare namespace wcl {
    const empty: {};
    const line: string;
    const logLine: string;

    function abort(): never;
    function isAbort(err: any): boolean;

    function pause(msg?: string): void;
    function readline(callBack: { (text: string): void }, msg?: string): void;

    function createFullPath(...subPath: string[]): Promise<string>;
    function appConfigPath(subPath: string, autoCreate?: boolean): Promise<string> | string;
    function appDataPath(subPath: string, autoCreate?: boolean): Promise<string> | string;
    function appTmpPath(subPath: string, autoCreate?: boolean): Promise<string> | string;

    //根据需要，自动创建带上下文调用的函数
    function fn(fnTarget: Function, context?: any): Function;

    //是否是对象
    function isObject(obj: any): boolean;

    //是否是数组
    function isArray(obj: any): boolean;

    //是否是函数
    function isFunction(obj: any): boolean;

    //是否是异步函数
    function isAsyncFunction(obj: any): boolean;

    //是否是字符串类型
    function isString(obj: any): boolean;

    //是否是数值类型
    function isNumber(obj: any): boolean;

    //返回所创建的新对象数组
    function makeObjArray(objDefine, valueArrayArray);

    // 将mergeFrom拥有的属性合并复制和覆盖到mergeTo里。
    function mergeObj(mergeFrom, mergeTo, onlyExists);

    /**
     * 将一列列对象的属性扩展到目标对象，并返回目标对象。    
     */
    function extends_(toObj?: object | any[], fromObj?: object | object[], options?: { onlyExists?: boolean, ignoreExists?: boolean }): object;
    //extends与typescript关键字冲突

    //转化为 prop1=value1;prop2=value2;...的形式
    function nvls(obj, spea);

    //返回对象的属性名称描述字符串
    function metaInfo(obj, withValue);

    //是否是空白字符串
    function isWhite(text: any): boolean;

    //是null时返回空字符串，否则返回text本身
    function notNull(text: any): string;

    //根据是否为空，自动连接字符串
    function merge(left: string, concat: string, right: string): string;

    //格式化数字
    function fmtNum(num: any, precision?: number, separator?: string): string;

    //格式化一个值
    function fmtValue(value: any, format?: string): string;

    //获取值的提示信息。func, obj, 或者toString()值。
    function valueInfo(obj);

    //获取对象的显示值。（为null,undefine时转换成空字符串）
    function prnOf(obj);

    //通过随机数生成一个新的guid
    const guid: string;

    const newId: number;

    function defConsts(sConst: string | object): object;

    //删除json中的 /*....*/ 注释部分
    function removeComment(srcJson: string): string;

    //将1段手写的json代码（最外面不包含{}）计算出结果后返回。
    function getData(text);

    //休眠指定的毫秒数（js为单线程）
    function sleep(millSeconds: number): void;

    function setLog(): void;

    function log(...msg: any[]): void;

    //返回当前时间。D: yyyy-MM-dd, T: hh:mm:ss, "L": yyyy-MM-dd hh:mm:ss, "D-": yyyyMMdd, "T-": hhmmss, "L-": yyyyMMdd hhmmss
    function now(format?: string): string;

    //使用 val(obj, "name1.name2.name3...")方式安全的访问对象的属性。确保不会因为undeinfe而报错
    //此方法可以单独拷贝到任何地方，不依赖其他函数。
    function val(obj: object, name: string): any;

    //以递归方式，获取一个对象的子属性名称列表
    function getDeepObjLayout(obj, maxDeep);

    /**
     * 获取第一个出现的路径扩展名。格式： [extName, start, end]
     */
    function getFirstExt(path: string, fullExtStyle?: boolean): string;

    /**
     * 获取路径中的第一段目录
     */
    function getFirstPath(path: string): string;

    //开启一个无限等待的循环, 直到call()为true或者到达了指定的最大尝试次数
    function wait(call?: { (times: number): void },
        option?: { ticks?: number, initTicks?: number, maxTimes?: number }): void;

    function joinPath(...args: string[]): string;

    /**
     * 返回destText与byText前导第一个不同的字符部分的索引
     */
    function getDiffIndex(byText: string, destText: string, filePathMode?: boolean): number;

    /**
     * 从destText剔除与byText前导相同的字符部分
     */
    function truncSameText(byText: string, destText: string,
        option: { string?: boolean, map?: string, filePathMode?: boolean }): number | string;

    /**
     * 格式化一个目录树列表
     */
    function formatDirTree(list: string[], sort?: boolean): string[];

    /**
     * 将一些类的实现混入到目标类（实现多继承）
     */
    function mixins(toClass: any, fromCalsses: any[]): any;

    function handleAppErr(): void;

    var toJson: {
        (obj: Object): string;
        option: { fmt: boolean };
    };
}

declare namespace wcl {
    namespace script {
        enum tokenType {
            string,
            int,
            float,
            identify,
            remark,
            symbol,
            whiteSpace,
            array,
            keyword,
        }

        class token {
            name: string;
            value: any;
            type: tokenType;

            constructor(theTokenType: tokenType, value?: any, name?: string);

            isSymbol(value?: string): boolean;
            isIdentify(name?: string): boolean;
            isIdentifyOrString(name?: string): boolean;
            isArray(): boolean;
            isKeyword(): boolean;
            getValue(): any;
            readonly isValue: boolean;
            toString(): string;
        }

        class reader {
            nameParameterSymbol: string;
            constructor(src: string, start?: number);
            readNext(): token;
            popNext(count?: number): token;
            peekNext(pos?: number): token;
        }

        function test(): void;
    }
}

declare namespace wcl {
    class stringBuilder {
        constructor();

        arr: string[];

        appendOne(val: string): stringBuilder;
        append(...val: string[]): stringBuilder;
        add(...val: string[]): stringBuilder;
        appendLine(...val: string[]): stringBuilder;
        back(count?: number): stringBuilder;
        backChar(count?: number): stringBuilder;
        clear(): stringBuilder;

        setCol(colObj: object): stringBuilder;
        fld(fldObj: object | string, val: any): stringBuilder;
        fld(fldObj: object | string, opr: string, val: any): stringBuilder;
        and(fldObj: object | string, val: any): stringBuilder;
        and(fldObj: object | string, opr: string, val: any): stringBuilder;
        andStart(fldObj: object | string, val: any): stringBuilder;
        or(fldObj: object | string, val: any): stringBuilder;
        or(fldObj: object | string, opr: string, val: any): stringBuilder;
        orStart(fldObj: object | string, val: any): stringBuilder;
        jEnd(): stringBuilder;

        toString(): string;
    }

    function sb(fldObj: object | string, val: any): stringBuilder;
    function sb(fldObj: object | string, opr: string, val: any): stringBuilder;
    function sb(...args: any[]): stringBuilder;
    function fld(fldObj: object | string, val: any): stringBuilder;
    function fld(fldObj: object | string, opr: string, val: any): stringBuilder;

    function fromCompactArray(compactArrayString: string): { ver: string, fields: string[], data: object[] };

    const char: {
        digit: RegExp,
        letter: RegExp,
        letterOrDigit: RegExp,
        symbol: RegExp,
        whiteSpace: RegExp,
    }

    class testTime {
        constructor();
        start(): void;
        mark(stepName?: string): void;
        useTime(): number;
        allUseTime(): string;
    }

    class cfgableString {
        static get(instObj: any, propertyKey: PropertyKey): any;
        constructor(fromString: string, option?: object);
    }

    const idHelper: {
        readonly newId: number;

        toId(val: number): number;
        rand(max?: number): number;
        time(id: number): Date;

        readonly MIN_ID: number;
    }
}

declare global {
    function log(...text: any[]): never;
    var _: _.LoDashStatic;
}

export = wcl;